class Message():
    """Class for passing messages.

    Messages in pykka must be passed as dictionaries. This class provides
    fields corresponding to the keys in that dictionary to facilitate the
    construction of such messages.
    """

    def __init__(self, text, metadata, should_log=False):
        self.text = text
        self.metadata = metadata
        self.should_log = should_log


class Metadata():
    """Class for passing Message metadata.

    Message Metadata is passed as a dictionary. This class provides
    fields corresponding to the keys in that dictionary to facilitate
    the construction of such dictionaries.
    """

    def __init__(self, source, dest=None, opts=None):
        self.source = source
        self.dest = dest
        self.opts = opts



class Error(Exception):
    """Base class for exception in the Legobot module

    """

    pass


class LegoError(Error):
    """Exceptions raised inside legos

    """

    def __init__(self, message):
        self.message = message
        Exception.__init__(self, message)



# Legobot
# Copyright (C) 2016 Brenton Briggs, Kevin McCabe, and Drew Bronson

import logging

from Legobot.Lego import Lego
from Legobot.Utilities import Utilities

logger = logging.getLogger(__name__)


class Help(Lego):
    @staticmethod
    def listening_for(message):
        if Utilities.isNotEmpty(message['text']):
            try:
                return message['text'].split()[0] == '!help'
            except Exception as e:
                logger.error(
                    'Help lego failed to check message text: {0!s}'.format(e))
                return False

    def handle(self, message):
        try:
            function = message['text'].split()[1]
        except IndexError:
            function = None

        baseplate_proxy = self.baseplate.proxy()
        legos = baseplate_proxy.children.get()
        lego_names = [
            lego.proxy().get_name().get()
            for lego in legos
            if lego.proxy().get_name().get()
            and lego.proxy().get_name().get() != '?'
        ]

        help_str = 'No help is available. Sorry.'

        if not function:
            help_str = 'Available functions: ' + ', '.join(lego_names)

        if function and function.lower() in [n.lower() for n in lego_names]:
            help_strs = []
            for lego in legos:
                name = lego.proxy().get_name().get()
                if name and name.lower() == function.lower():
                    try:
                        sub = message['text'].split()[2]
                        try:
                            temp = lego.proxy().get_help(sub=sub).get()
                        except (TypeError, KeyError):
                            temp = '{} has no information on {}'.format(
                                function, sub)
                    except IndexError:
                        temp = lego.proxy().get_help().get()

                    help_strs.append(temp)

            help_str = '\n'.join(help_strs)

        opts = self.build_reply_opts(message)

        self.reply(message, help_str, opts=opts)

    @staticmethod
    def get_name():
        return None



import json
import logging
import threading
import time

import jmespath
from pykka import ActorRegistry
from pykka import ThreadingActor

from Legobot.LegoError import LegoError
from Legobot.Message import Message
from Legobot.Message import Metadata

logger = logging.getLogger(__name__)


class Lego(ThreadingActor):
    class HandlerThread(threading.Thread):
        """This class provides a simple thread for running message handlers.

        It is used to ensure that message handlers do not block other Legos
        from running by simply taking too long to execute.
        """

        def __init__(self, handler, message):
            threading.Thread.__init__(self)
            self.handler = handler
            self.message = message

        def run(self):
            self.handler(self.message)

    def __init__(self, baseplate, lock: threading.Lock, log_file=None,
                 acl=None, rate_config=None):
        """Initializes the thread

        :param baseplate: the baseplate Lego, which should be \
                          the same instance of Lego for all Legos
        :param lock: a threading lock, which should be the same \
                     instance of threading.Lock for all Legos
        :param log_file str: a file path for writing logs to file
        :param acl dict: a dict representing an access control list. \
                         ex: {'whitelist': [list of source_users]}
        :param rate_config dict: a dict representing the rate limite config. \
                                 ex: {'rate_key': jmespath expr to get rate \
                                        key value from message object, \
                                      'rate_interval': the rate limit \
                                        interval in seconds}
        """
        super().__init__()
        if not lock:
            raise LegoError("Lock expected but not provided!")
        self.baseplate = baseplate
        self.children = []
        self.lock = lock
        self.log_file = log_file
        self.acl = acl if acl else {}

        # set rate limit items
        self.set_rate_limit(rate_config)

    def set_rate_limit(self, rate_config):
        """Set rate limit config for this Lego.

        :param rate_config dict: dict representing the rate limit config
        """
        rate_config = rate_config if isinstance(rate_config, dict) else {}
        self.rate_key = rate_config.get('rate_key')
        self.rate_interval = rate_config.get('rate_interval')
        self.rate_log = {}

    def on_receive(self, message):
        """Handle being informed of a message.

        This function is called whenever a Lego receives a message, as
        specified in the pykka documentation.

        Legos should not override this function.

        :param message:
        :return:
        """
        if self.log_file is not None and message['should_log']:
            message_copy = Message(message['text'],
                                   Metadata(None).__dict__,
                                   message['should_log']).__dict__
            with open(self.log_file, mode='w') as f:
                f.write(json.dumps(message_copy))
            logger.info(message['metadata']['source'])
        if (self.acl_check(message)
                and self.rate_check(message)
                and self.listening_for(message)):
            self_thread = self.HandlerThread(self.handle, message)
            self_thread.start()
        self.cleanup()
        for child in self.children:
            child.tell(message)

    def cleanup(self):
        """Clean up finished children.

        :return: None
        """
        self.lock.acquire()
        logger.debug('Acquired lock in cleanup for ' + str(self))
        self.children = [child for child in self.children if child.is_alive()]
        self.lock.release()

    def rate_check(self, message):
        """Return whether the message passes the rate limit check for this Lego

        :param message: a Message object
        :return: Boolean
        """
        if not self.rate_key or not self.rate_interval:
            return True

        now = int(round(time.time()))
        key = jmespath.search(self.rate_key, message)
        last_invoke = self.rate_log.get(key, 0)

        if now - last_invoke >= self.rate_interval:
            self.rate_log[key] = now
            return True
        else:
            return False

    def acl_check(self, message):
        """Return whether the message passes the ACL check for this Lego.

        :param message: a Message object
        :return: Boolean
        """
        user = message.get('metadata', {}).get('source_user')
        acl_conditions = [
            user,
            isinstance(user, str),
            self.acl,
            isinstance(self.acl, dict)
        ]
        if all(acl_conditions):
            whitelist = self.acl.get('whitelist', [])
            if whitelist and user in whitelist:
                return True
            elif whitelist and user not in whitelist:
                return False

            blacklist = self.acl.get('blacklist', [])
            if blacklist and user in blacklist:
                return False
            if blacklist and user not in blacklist:
                return True

        return True

    def listening_for(self, message):
        """Return whether this Lego is listening for the provided Message.

        All Legos should override this function.

        :param message: a Message object
        :return: a boolean
        """
        return False

    def handle(self, message):
        """Handle the provided Message.

        All Legos should override this function.

        :param message: a Message object
        :return: None
        """
        return

    def add_child(self, child_type, *args, **kwargs):
        """Initialize and keep track of a child.

        :param child_type: a class inheriting from Lego to initialize \
                           an instance of
        :param args: arguments for initializing the child
        :param kwargs: keyword arguments for initializing the child
        :return:
        """
        try:
            baseplate = kwargs['baseplate']
        except Exception:
            if self.baseplate is None:
                baseplate = self.actor_ref
            else:
                baseplate = self.baseplate
        try:
            lock = kwargs['lock']
        except Exception:
            lock = self.lock
        child = child_type.start(baseplate, lock, *args, **kwargs)
        self.children.append(child)

    def reply(self, message, text, opts=None):
        """Reply to the sender of the message with a containing the text

        :param message: the message to reply to
        :param text: the text to reply with
        :param opts: A dictionary of additional values to add to metadata
        :return: None
        """
        metadata = Metadata(source=self.actor_urn,
                            dest=message['metadata']['source']).__dict__
        metadata['opts'] = opts
        message = Message(text=text, metadata=metadata,
                          should_log=message['should_log']).__dict__
        dest_actor = ActorRegistry.get_by_urn(message['metadata']['dest'])
        if dest_actor is not None:
            dest_actor.tell(message)
        else:
            raise("Tried to send message to nonexistent actor")

    def reply_attachment(self, message, text, attachment, opts=None):
        """Convenience method for formatting reply as attachment (if available)

        Passes attachment on to the reply method. Individual connectors can
        deal with the attachment or simply pass it on as a regular message

        :param message: the message to reply to
        :param text: the text to reply with
        :param attachment: the attachment link
        :param opts: A dictionary of additional values to add to metadata
        :return: None
        """
        if not opts:
            opts = {}

        opts['attachment'] = attachment
        opts['fallback'] = text
        text += '\n {}'.format(attachment)
        self.reply(message, text, opts)

    def build_reply_opts(self, message):
        """Convenience method for constructing default options for a reply

        :param message: the message to reply to
        :return: opts
        """
        try:
            source = message['metadata']['source_channel']
            thread = message['metadata'].get('thread_ts')
            opts = {'target': source, 'thread': thread}
        except LookupError:
            source = None
            opts = None
            logger.error("Could not identify source from  message:{}\n"
                         .format(str(message)))
        return opts

    def get_name(self):
        """Return the name the Lego recognizes from the help function.

        :return: a string
        """
        return '?'

    def get_help(self):
        """Return a helpstring for the function.

        :return: a string
        """
        return ''

    def on_failure(self, exception_type, exception_value, traceback):
        ref = ActorRegistry.get_by_urn(self.actor_urn)
        logger.exception('Lego crashed: ' + str(ref))
        logger.exception(exception_type)
        logger.exception(exception_value)



import logging
from pathlib import Path
from pydoc import locate
import threading

from jsonschema import validate
import yaml

from Legobot.Lego import Lego


DIR = Path(__file__).resolve().parent
HELP_PATH = 'Legobot.Legos.Help.Help'


def build_logger(log_file=None, level=None):
    if (
        not level
        or not isinstance(level, str)
        or level.upper() not in ['INFO', 'DEBUG', 'ERROR']
    ):
        level = 'DEBUG'

    level = getattr(logging, level.upper())

    if log_file:
        logging.basicConfig(filename=log_file)

    logger = logging.getLogger()
    logger.setLevel(level)
    ch = logging.StreamHandler()
    ch.setLevel(level)

    # create formatter and add it to the handlers
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)

    # add the handlers to the logger
    logger.addHandler(ch)

    return logger


def load_yaml_file(path, default=None, raise_ex=None, logger=None):
    if not logger:
        logger = build_logger()

    if not isinstance(path, Path):
        path = Path(path)

    try:
        with path.open() as f:
            out = yaml.safe_load(f)

        logger.debug(f'Loaded {path} successfully.')
    except Exception as e:
        out = default
        if raise_ex is True:
            raise(e)
        else:
            logger.error(f'Error loading {path}:\n {e}\nReturning default.')

    return out


class Chatbot(object):
    def __init__(self, config_path=None):
        self.schema = load_yaml_file(
            DIR.joinpath('chatbot_schema.yaml'), raise_ex=True)
        self.config = self.load_config(config_path)
        self.logger = build_logger(
            self.config.get('log_file'), self.config.get('log_level'))
        self.baseplate = None
        self.baseplate_proxy = None
        self.connectors = []
        self.legos = []

    def load_config(self, config_path):
        config = load_yaml_file(config_path, raise_ex=True)
        validate(config, self.schema)
        return config

    def initialize_baseplate(self):
        lock = threading.Lock()
        self.baseplate = Lego.start(None, lock)
        self.baseplate_proxy = self.baseplate.proxy()

    def add_lego(self, name, config, lego_type):
        if config['enabled'] is True:
            try:
                self.baseplate_proxy.add_child(
                    locate(config['path']), **config.get('kwargs', {}))

                if lego_type == 'connectors':
                    self.connectors.append(name)
                elif lego_type == 'legos':
                    self.legos.append(name)

            except Exception as e:
                self.logger.error(f'Error adding {name} to {lego_type}: {e}')

    def run(self):
        if not self.baseplate:
            self.initialize_baseplate()

        if self.config['helpEnabled'] is True:
            self.add_lego(
                'Help', {'enabled': True, 'path': HELP_PATH}, 'legos')

        for connector, config in self.config['connectors'].items():
            self.add_lego(connector, config, 'connectors')

        for lego, config in self.config['legos'].items():
            self.add_lego(lego, config, 'legos')

    def stop(self):
        self.baseplate.stop()
        self.baseplate_proxy = None
        self.baseplate = None
